from flask import Blueprint, jsonify, request
from flask_login import current_user

from stumpf.models import (AnnotationApp, AnnotationAppSchema, Artifact,
                           ArtifactSchema, Assignment, AssignmentSchema,
                           db, Task, TaskSchema, Token, TokenSchema, UserSchema,
                           Video, VideoSchema)
from stumpf.pagination import paginated_response
from stumpf.sorting import sorted_query

annotation_app_bp = Blueprint('annotation_app', __name__)
artifact_bp = Blueprint('artifact', __name__)
video_bp = Blueprint('video', __name__)
assignment_bp = Blueprint('assignment', __name__)
task_bp = Blueprint('task', __name__)
token_bp = Blueprint('token', __name__)
user_bp = Blueprint('user', __name__)

annotation_app_schema = AnnotationAppSchema()
artifact_schema = ArtifactSchema()
assignment_schema = AssignmentSchema()
task_schema = TaskSchema()
token_schema = TokenSchema()
user_schema = UserSchema()
video_schema = VideoSchema()


@annotation_app_bp.route('/annotation-app', methods=['POST'])
def create_annotation_app():
    try:
        annotation_app = annotation_app_schema.load(request.json)

        if AnnotationApp.query.filter_by(name=annotation_app.name,
                                         repo_url=annotation_app.repo_url,
                                         url=annotation_app.url,
                                         hash=annotation_app.hash).first():
            return jsonify({
                'error': 'Annotation app already exists'
            }), 409

        db.session.add(annotation_app)
        db.session.commit()
        return jsonify(annotation_app_schema.dump(annotation_app)), 201
    except Exception as e:
        db.session.rollback()
        raise


@annotation_app_bp.route('/annotation-app', methods=['GET'])
def list_annotation_app():
    query = sorted_query(AnnotationApp)
    return paginated_response(query, lambda items: annotation_app_schema.dump(items, many=True))


@annotation_app_bp.route('/annotation-app/<uuid:annotation_app_id>', methods=['GET'])
def get_annotation_app(annotation_app_id):
    annotation_app = AnnotationApp.query.get_or_404(annotation_app_id)
    return jsonify(annotation_app_schema.dump(annotation_app))


@annotation_app_bp.route('/annotation-app/<uuid:annotation_app_id>', methods=['DELETE'])
def delete_annotation_app(annotation_app_id):
    try:
        AnnotationApp.query.filter_by(id=annotation_app_id).delete()
        db.session.commit()
        return '', 204

    except Exception as e:
        db.session.rollback()
        raise


@user_bp.route('/user/me')
def get_current_user():
    return jsonify(user_schema.dump(current_user))


@token_bp.route('/token', methods=['POST'])
def create_token():
    token_json = request.json
    token_json['user_id'] = current_user.get_id()
    token = token_schema.load(token_json)
    db.session.add(token)
    db.session.commit()
    return jsonify(token_schema.dump(token))


@token_bp.route('/token/<token_value>')
def get_token(token_value):
    token = Token.query.filter_by(token=token_value).first_or_404()
    return jsonify(token_schema.dump(token))


@token_bp.route('/token/<token_value>', methods=['DELETE'])
def delete_token(token_value):
    token = Token.query.filter_by(token=token_value).first_or_404()
    db.session.delete(token)
    db.session.commit()
    return '', 204


@video_bp.route('/video', methods=['POST'])
def create_video():
    try:
        video = video_schema.load(request.json)

        if Video.query.filter_by(path=video.path).first():
            return jsonify({
                'error': 'Video path already exists'
            }), 409
        elif Video.query.filter_by(name=video.name).first():
            return jsonify({
                'error': 'Video name already exists'
            }), 409

        db.session.add(video)
        db.session.commit()
        return jsonify(video_schema.dump(video)), 201
    except Exception as e:
        db.session.rollback()
        raise

    return jsonify(video_schema.dump(video))


@video_bp.route('/video', methods=['GET'])
def list_videos():
    query = sorted_query(Video)
    return paginated_response(query, lambda items: video_schema.dump(items, many=True))


@video_bp.route('/video/<uuid:video_id>', methods=['GET'])
def get_video(video_id):
    video = Video.query.get_or_404(video_id)
    return jsonify(video_schema.dump(video))


@task_bp.route('/task', methods=['POST'])
def create_task():
    try:
        task = task_schema.load(request.json)
        Video.query.get_or_404(task.video_id)

        if Task.query.filter_by(name=task.name).first():
            return jsonify({
                'error': 'Task with that name already exists'
            }), 409

        db.session.add(task)
        db.session.commit()
        return jsonify(task_schema.dump(task)), 201
    except Exception as e:
        db.session.rollback()
        raise

    return jsonify(task_schema.dump(task))


@task_bp.route('/task', methods=['GET'])
def list_tasks():
    query = sorted_query(Task)
    return paginated_response(query, lambda items: task_schema.dump(items, many=True))


@assignment_bp.route('/assignment', methods=['GET'])
def list_assignments():
    query = sorted_query(Assignment)
    return paginated_response(query, lambda items: assignment_schema.dump(items, many=True))


@assignment_bp.route('/assignment', methods=['POST'])
def create_assignment():
    try:
        assignment = assignment_schema.load(request.json)

        if not Task.query.get(assignment.task_id):
            return jsonify({
                'error': 'No such task.'
            }), 400
        elif not Video.query.get(assignment.video_id):
            return jsonify({
                'error': 'No such video.'
            }), 400

        db.session.add(assignment)
        db.session.commit()
        return jsonify(assignment_schema.dump(assignment)), 201
    except Exception as e:
        db.session.rollback()
        raise

    return jsonify(assignment_schema.dump(assignment))


# todo put vs patch
@assignment_bp.route('/assignment/<uuid:assignment_id>', methods=['PUT'])
def update_assignment(assignment_id):
    assignment = Assignment.query.get_or_404(assignment_id)

    if assignment.completed is not None:
        return jsonify({
            'error': 'Assignment is already completed'
        }), 409

    try:
        assignment_update_attrs = assignment_schema.load(request.json, only=('completed',))

        for attr, value in assignment_update_attrs.items():
            setattr(assignment, attr, value)

        db.session.add(assignment)
        db.session.commit()
        return jsonify(assignment_schema.dump(assignment)), 201
    except Exception as e:
        db.session.rollback()
        raise

    return jsonify(assignment_schema.dump(assignment))


@artifact_bp.route('/artifact', methods=['POST'])
def create_artifact():
    try:
        artifact = artifact_schema.load(request.json)
        assignment = Assignment.query.get(artifact.assignment_id)

        if Artifact.query.filter_by(path=artifact.path).first():
            return jsonify({
                'error': 'Artifact with that path already exists'
            }), 409
        elif not assignment:
            return jsonify({
                'error': 'Assignment does not exist'
            }), 400
        elif assignment.completed is not None:
            return jsonify({
                'error': 'Assignment has already been completed'
            }), 409

        db.session.add(artifact)
        db.session.commit()
        return jsonify(artifact_schema.dump(artifact)), 201
    except Exception as e:
        db.session.rollback()
        raise

    return jsonify(artifact_schema.dump(artifact))


@artifact_bp.route('/artifact/<uuid:assignment_id>', methods=['GET'])
def list_artifacts(assignment_id):
    Assignment.query.get_or_404(assignment_id)
    query = Artifact.query.filter_by(assignment_id=assignment_id)
    query = sorted_query(Artifact, query=query)
    return paginated_response(query, lambda items: artifact_schema.dump(items, many=True))
